#ifndef FITNESS_H
#define FITNESS_H

#include <eo>
#include <iostream> 
#include <fstream>
#include <vector>

#include "dataset.h" 

using namespace std;


class Keywords : public EO<double> {

public:

  
  Vector operator[](int i) const { return centers[i]; } 
  Vector& operator[](int i)  { return centers[i]; } 

  void resize(int N) { centers.resize(N);  }
  unsigned size() const { return centers.size(); }
  
  void remove(int i) { centers[i] = centers[ centers.size()-1 ]; centers.pop_back(); } 
  void add(Vector& x) { centers.push_back(x); } 

  virtual std::string className() const {
    return "Keywords";
  }
 
  double  nearestDist(const Vector& x) { // distance from x to the closest center 
    double dist = centers[0].distance(x); 
    double d; 
    for (unsigned i=1; i<centers.size(); i++) {
      d = centers[i].distance2(x);
      if (d < dist) 
	dist = d; 
    }
    return dist; 
  } 

  int  nearest(const Vector& x) { // distance from x to the closest center 
    int index = 0; 
    double dist = centers[0].distance(x); 
    double d; 
    for (unsigned i=1; i<centers.size(); i++) {
      d = centers[i].distance(x);
      if (d < dist) { 
	dist = d; 
	index = i; 
      }
    }
    return index; 
  } 


  virtual void printOn(std::ostream& os) const {
    EO<double>::printOn(os);
    os << "[" << centers.size()  << "]" << endl; 
    /*
    for (unsigned i=0; i<centers.size(); i++) {
      for (unsigned j=0; j<centers[i].size(); j++) 
	os << centers[i][j] << " ";
      os << endl; 
    }
    */
      
  }
  
  virtual void readFrom(std::istream& is)   {
    // not implemented
  }

  //private:
  vector<Vector> centers;
  vector<int> counters;
}; 



// FITNESS FUNCTION 
// father of my fitness functions 
// normalize method 
// 
class MyFitness : public eoEvalFunc<Keywords> {

public: 
  MyFitness(DataSet& _data) : data(_data) {};

  // moves the keywords to the centers of coresponding clusters
  void normalize(Keywords& ind) {
    // one step of k-means
    
    // setup 
    ind.counters.resize( ind.size() );
    for (unsigned i=0; i<ind.counters.size(); i++) ind.counters[i] = 0;
    
    centers.resize( ind.size() );
    for (unsigned i=0; i<centers.size(); i++) centers[i].zero( data.getN() ); 

    // go through training set and assign training points to clusters
    int index; 
    for (int i=0; i<data.size(); i++) {
      index = ind.nearest( data[i].input ); 
      ind.counters[ index ]++;
      centers[index] += data[i].input; 
    } 
    
    // comp means and move centers 
    for (unsigned i=0; i<centers.size(); i++) 
      if (ind.counters[i] > 0) 
	centers[i] /= ind.counters[i]; 

    ind.centers = centers; 
  };  


protected:
  // vector<int> counters;
  vector<Vector> centers;

  //private:
  DataSet& data; 
}; 

class SumDistancesFitness :  public MyFitness {

public:

  SumDistancesFitness(DataSet& _data) : MyFitness(_data) {
  };
   
  
  virtual void operator() (Keywords& ind)   {
    

    if (ind.invalid()) {      

      // cout << "Comp. fitness" << endl;
      
      normalize(ind); 
      
      // fitness comp 
      double dist = 0;
          
      for (int i=0; i<data.size(); i++) 
	dist += ind.nearestDist(data[i].input);

      
      dist /= data.size();
      dist *= 100;
      ind.fitness( - dist );       
    };
    
  };
};


class SilhouetteFitness :  public MyFitness {

public:

  SilhouetteFitness(DataSet& _data) : MyFitness(_data) {
  };
  
  virtual void operator() (Keywords& ind)   {
    if (ind.invalid()) {      

      // dopiste vypocet fitness
      
      // ind.fitness( spocitana silueta ); 
    }
  }; 

}; 

#endif
